import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Scanner;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.TreeSet;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.stream.Stream;

public class ParallelMTIS {

	public static void main(String[] args) {
		int[][] grid = readGrid(args[0]);
		ParallelMTIS mtis = new ParallelMTIS(grid);
		Set<Vertex> mis = mtis.findMTIS();
		for (Vertex v : mis)
			System.out.println(v);
		
	}

	private int verticesCount;
	private Set<Vertex> vertices;
	private Set<Edge> edges;
	private Set<Vertex> mis;
	private int[][] adjmatrix;
	
	public ParallelMTIS(int[][] grid) {
		
		verticesCount = grid[0].length;
		vertices = new HashSet<Vertex>();
		edges = new HashSet<Edge>();
		mis = new HashSet<Vertex>();
		this.adjmatrix = grid;
		for (int i = 0; i < verticesCount; i++) {
			Vertex v = new Vertex(i);
			for (int j = 0; j < verticesCount; j++) {
				//System.out.print(grid[i][j] + " ");
				
				if (grid[i][j] == 1) {
					edges.add(new Edge(i,j));
					v.degrees++;
				}
			}
			//System.out.println();
			vertices.add(v);
		}
	}
	
	public Set<Vertex> findMTIS() {
		
		while (!vertices.isEmpty()) {
			markVertices();
			breakTies();
			findMaximal();
			//System.out.println(vertices.size());
		}
		return mis;
	}
	
	/*
	 * Mark each vertex in parallel
	 */
	private void markVertices() {
		
		List<Vertex> v = new ArrayList<Vertex>(vertices);
		MarkVertices vm = new MarkVertices(v);
		vm.start();
		vm.sync();
	}
	
	private void breakTies() {
		List<Vertex> v = new ArrayList<Vertex>(vertices);
		List<Edge> e = new ArrayList<Edge>(edges);
		BreakEdgeTies bet = new BreakEdgeTies(v, e, adjmatrix);
		bet.start();
		bet.sync();
	}
	
	private void findMaximal() {
		List<Vertex> v = new ArrayList<Vertex>(vertices);
		List<Vertex> output = new ArrayList<Vertex>();
		MaximalFinder finder = new MaximalFinder(v, output);
		finder.start();
		finder.sync();
		mis.addAll(output);
		/*for (Vertex d: mis)
			System.out.println("mis "+d.id);*/
		removeNeighbors(output);
	}
	
	/*
	 * Remove the marked node v and gamma(v) (its neighbors) if
	 * they reside in a triangle.
	 *  (Don't remove if they aren't in a triangle)
	 *  
	 *  Also remove edges connecting neighbors to other nodes
	 * 
	 */
	private void removeNeighbors( List<Vertex> vertex) {
		Set<Vertex> neighborVertices = new HashSet<Vertex>();
		List<Edge> neighborEdges = new ArrayList<Edge>();
		List<Vertex> tempVertices = new ArrayList<Vertex>(vertices);
		for (Vertex v: vertex) {
			List<Edge> tempEdges = new ArrayList<Edge>(edges);

			//get all edges with neighbors
			for (Edge e : tempEdges) 
				if (e.contains(v)) 
					neighborEdges.add(e);
				

			//find triangles here
			for (Edge neighborEdge: neighborEdges) {
				
				
				for (int i = 0; i < verticesCount; i++) {
    				if (adjmatrix[neighborEdge.end1][i] == 1 && i != neighborEdge.end2) { 
    					if (adjmatrix[neighborEdge.end2][i] == 1) {
    						
    						if (tempVertices.contains((new Vertex(i))))
								neighborVertices.add(tempVertices.get(tempVertices.indexOf(new Vertex(i))));
    						
    						if (tempVertices.contains((new Vertex(neighborEdge.end1))))
								neighborVertices.add(tempVertices.get(tempVertices.indexOf(new Vertex(neighborEdge.end1))));
    						
    						if (tempVertices.contains((new Vertex(neighborEdge.end2))))
								neighborVertices.add(tempVertices.get(tempVertices.indexOf(new Vertex(neighborEdge.end2))));
    						
    					}
    				
    				}
    			}
				
			}
			
		}
		
		//remove the lingering edges of the neighboring nodes
		for (Vertex v: neighborVertices) {
			List<Edge> tempEdges = new ArrayList<Edge>(edges);

			for (Edge e : tempEdges) {
				if (e.contains(v)) {
					neighborEdges.add(e);
					Vertex v2 = null;
					
					if (v.id != e.end1) {
						v2 = new Vertex(e.end1);
						v2.degrees = tempVertices.get(tempVertices.indexOf(v2)).degrees - 1;

					}
					else if (v.id != e.end2) {
						v2 = new Vertex(e.end2);
						v2.degrees = tempVertices.get(tempVertices.indexOf(v2)).degrees - 1;

					}
					
					if (v2 != null && !neighborVertices.contains(v2)) {
						vertices.remove(v2);
						vertices.add(v2);
					}
				}
			}
		}
		
		vertices.removeAll(vertex);
		vertices.removeAll(neighborVertices);
		edges.removeAll(neighborEdges);
	};
	
    public static int[][] readGrid(String fileName) {
    	
    	try {
    		File input = new File(fileName);
			Scanner scanner = new Scanner(input);
			String tempGrid = "";
			int rows = 0;
			int cols = 0;
			while (scanner.hasNextLine()) {
				String lineToParse = scanner.nextLine();
				if (lineToParse.trim().isEmpty())
					continue;
				//String tempLine = scanner.nextLine();
				StringTokenizer tempTokenizer = new StringTokenizer(lineToParse);

				cols = tempTokenizer.countTokens();
				tempGrid += lineToParse + " ";
				rows++;
			}
			scanner.close();
			
			int[][] grid = new int[rows][cols];
			StringTokenizer tokenizer = new StringTokenizer(tempGrid);
			for (int i = 0; i < rows; i++) {
				for (int j = 0; j < cols; j++) {
					grid[i][j] = Integer.parseInt(tokenizer.nextToken());
				}
				
			}
			return grid;
			
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
        return null;

    }
    
    /*
     * The class will recursively generate thread for each vertex 
     * and then randomly pick to mark the vertex by odds 1/2d(v)
     */
    class MarkVertices extends Thread {
    	List<Vertex> vertices;
    	boolean done = false;
    	public MarkVertices(List<Vertex> v) {
    		this.vertices = v;
    	}
    	
    	public void run() {
    		
    		if (vertices.size() == 1) {
	    		Vertex v = vertices.get(0);
	    		if (v.degrees == 0) {
	    			mis.add(v);
	    			return;
	    		}
	    		Random r = new Random();
	    		if (r.nextInt(2*v.degrees) == 0) {
	    			v.setMark(true);
	    			//System.out.println(v.id + " marked");
	    		}

    		} else {
    			List<Vertex> half1 = new ArrayList<Vertex>(vertices.subList(0, vertices.size()/2));
    			List<Vertex> half2 = new ArrayList<Vertex>(vertices.subList(vertices.size()/2, vertices.size()));
    			MarkVertices a, b;
    			a = new MarkVertices(half1);
    			b = new MarkVertices(half2);

    			a.start();
    			b.start();

    			a.sync();
    			b.sync();
    		}
			done = true;

    	}
    	
    	// Professor's code
        public synchronized void sync()
        {
            if(!done) {
                try {
                    wait(); 
                } catch(InterruptedException ie) {
                    ie.printStackTrace();
                }
            }
        }
    	
    }
    
    /*
     * Recursively generate threads for each edge and 
     * break any ties if both vertices are marked.
     * 
     * Ties are broken by lower degree count then by lower id
     */
    class BreakEdgeTies extends Thread {
    	List<Vertex> vertices;
    	List<Edge> edges;
    	int[][] grid;
    	boolean done = false;

    	public BreakEdgeTies(List<Vertex> vertices, List<Edge> edge, int[][] grid) {
    		this.vertices = vertices;
    		this.edges = edge;
    		this.grid = grid;
    	}
    	
    	public void run() {
    		
    		if (edges.size() == 1) {
    			
	    		Edge e = edges.get(0);
	    		Vertex v1 = vertices.get(vertices.indexOf(new Vertex(e.end1)));
	    		Vertex v2 = vertices.get(vertices.indexOf(new Vertex(e.end2)));
	    		if (v1.getMark() && v2.getMark()) {
	    				    			
	    			//select a marked vertex and look for neighbors excluding v2
	    			for (int i = 0; i < verticesCount; i++) {
	    				if (grid[v1.id][i] == 1 && i != v2.id) { 
	    					if (grid[v2.id][i] == 1) {
	    						
	    						//unmark by lower degree, then by id
	    		    			if (v1.degrees == v2.degrees) {
	    		    				if (v1.id < v2.id)
	    		    					v1.setMark(false);
	    		    				else
	    		    					v2.setMark(false);
	    		    			} else if (v1.degrees < v2.degrees)
	    		    				v1.setMark(false);
	    		    			else
	    		    				v2.setMark(false);
	    						break;
	    					}
	    				
	    				}
	    			}
	    			

	    		}

    		} else if (edges.size() > 1 ){

    			
    			List<Edge> half1 = new ArrayList<Edge>(edges.subList(0, edges.size()/2));
    			List<Edge> half2 = new ArrayList<Edge>(edges.subList(edges.size()/2, edges.size()));
    			BreakEdgeTies a, b;
    			a = new BreakEdgeTies(vertices, half1, grid);
    			b = new BreakEdgeTies(vertices, half2, grid);

    			a.start();
    			b.start();
    			
    			a.sync();
    			b.sync();
    		}
			done = true;

    	}
    	
    	// Professor's code
        public synchronized void sync()
        {
            if(!done) {
                try {
                    wait(); 
                } catch(InterruptedException ie) {
                    ie.printStackTrace();
                }
            }
        }
    	
    	
    }
    
    /*
     * Adds the vertices to the output list in parallel
     */
    class MaximalFinder extends Thread {
		List<Vertex> vertices;
		List<Vertex> s; //output
		boolean done = false;
    	public MaximalFinder(List<Vertex> vertices, List<Vertex> output) {
    		this.vertices = vertices;
    		this.s = output; 
    	}
    	
    	public void run() {
    		
    		if (vertices.size() == 1 ) {

    			if( vertices.get(0).getMark())
    				s.add(vertices.get(0));
    		} else {
    			List<Vertex> half1 = new ArrayList<Vertex>(vertices.subList(0, vertices.size()/2));
    			List<Vertex> half2 = new ArrayList<Vertex>(vertices.subList(vertices.size()/2, vertices.size()));
    			MaximalFinder a, b;
    			a = new MaximalFinder(half1, s);
    			b = new MaximalFinder(half2, s);
    			
    			a.start();
    			b.start();
    			
    			a.sync();
    			b.sync();
    			
    		}
			done = true;

    	}
    	
    	// Professor's code
        public synchronized void sync()
        {
            if(!done) {
                try {
                    wait(); 
                } catch(InterruptedException ie) {
                    ie.printStackTrace();
                }
            }
        }
    }
    
    // in hindsight I could do away with the edge class and have each
    // vertex track its neighbors individually, but I do not have time to rewrite
    class Vertex {
    	int id;
    	int degrees;
    	AtomicBoolean marked;
    	AtomicBoolean neverMark;
    	public Vertex(int id) {
    		this.id = id;
    		this.marked = new AtomicBoolean(false);
    		this.neverMark = new AtomicBoolean(false);
    		this.degrees = 0;
    	}
    	
    	public void setMark(boolean mark) {
    		marked.set(mark);
    	}
    	
    	
    	public boolean getMark() {
    		return marked.get();
    	}
    	
    	@Override
    	public boolean equals(Object o) {
			return o instanceof Vertex && ((Vertex) o).id == id;
    	}
    	
    	@Override
    	public int hashCode() {
    	    return id;
    	}
    	
    	@Override
    	public String toString() {
    		return "id: " + id + " deg: " + degrees + " marked: " + marked;
    	}
    	
    }
    
    class Edge {
    	int end1;
    	int end2;
    	
    	public Edge(int end1, int end2) {
    		this.end1 = end1;
    		this.end2 = end2;
    	}
    	
    	public boolean contains(Vertex v) {
    		try {
    		if (end1 == v.id || end2 == v.id)
    			return true;
    		} catch (Exception e) {
    			e.printStackTrace();
    		}
    		return false;

    	}
    	
    	@Override
    	public boolean equals(Object o) {
			return o instanceof Edge && 
					((((Edge) o).end1 == end1 && ((Edge) o).end2 == end2) ||
					(((Edge) o).end1 == end2 && ((Edge) o).end2 == end1));
    	}
    	
    	@Override
    	public int hashCode() {
    	    int hash = 7;
    	    hash = 71 * hash + this.end1 + this.end2;
    	    hash = 71 * hash + (this.end1 * this.end2);
    	    return hash;
    	}
    	
    	@Override
    	public String toString() {
    		return  end1 + "-" + end2;
    	}
    }
    
}
